from base64 import b64encode
import hashlib
import os
import re
from sys import maxsize as MAXSIZE
from typing import (
    Dict,
    IO,
    Optional,
    TYPE_CHECKING,
)

import requests

from .exceptions import TusCommunicationError
from .request import catch_requests_error
# from .fingerprint import *

# from .storage_interface import Storage

if TYPE_CHECKING:
    from .client import TusClient


class BaseUploader:
    """
    Object to control upload related functions.

    :Attributes:
        - file_path (str):
            This is the path(absolute/relative) to the file that is intended
            for upload
            to the tus server. On instantiation this attribute is required.
        - file_stream (file):
            As an alternative to the `file_path`, an instance of the
            file
            to be uploaded
            can be passed to the constructor as `file_stream`. Do note that
            either the
            `file_stream` or the `file_path` must be passed on instantiation.
        -  url (str):
            If the upload url for the file is known, it can be passed to the
            constructor.
            This may happen when you resume an upload.
        - client (<tusclient.client.TusClient>):
            An instance of `tusclient.client.TusClient`. This would tell the
            uploader instance
            what client it is operating with. Although this argument is
            optional, it is only
            optional if the 'url' argument is specified.
        - chunk_size (int):
            This tells the uploader what chunk size(in bytes) should be
            uploaded when the method `upload_chunk` is called.
            This defaults to the maximum possible integer if not
            specified.
        - metadata (dict):
            A dictionary containing the upload-metadata. This would be encoded
            internally
            by the method `encode_metadata` to conform with the tus protocol.
        - offset (int):
            The offset value of the upload indicates the current position of
            the file upload.
        - stop_at (int):
            At what offset value the upload should stop.
        - request (<tusclient.request.TusRequest>):
            A http Request instance of the last chunk uploaded.
        - retries (int):
            The number of attempts the uploader should make in the case of a
            failed upload.
            If not specified, it defaults to 0.
        - retry_delay (int):
            How long (in seconds) the uploader should wait before retrying a
            failed upload attempt.
            If not specified, it defaults to 30.
        - store_url (bool):
            Determines whether or not url should be stored, and uploads should be
            resumed.
        - url_storage (<tusclient.storage.interface.Storage>):
            An implementation of <tusclient.storage.interface.Storage> which is
            an API for URL storage.
            This value must be set if store_url is set to true. A ready to use
            implementation exists atbe used out of the box. But you can
            implement your own custom storage API and pass an instace of it as
            value.
        - fingerprinter (<tusclient.fingerprint.interface.Fingerprint>):
            An implementation of <tusclient.fingerprint.interface.Fingerprint>
            which is an API to generate
            a unique fingerprint for the uploaded file. This is used for url
            storage when resumability is enabled.
            if store_url is set to true, the default fingerprint module
            (<tusclient.fingerprint.fingerprint.Fingerprint>)
            would be used. But you can set your own custom fingerprint module by
            passing it to the constructor.
        - upload_checksum (bool):
            Whether or not to supply the Upload-Checksum header along with each
            chunk. Defaults to False.

    :Constructor Args:
        - file_path (str)
        - file_stream (Optional[file])
        - url (Optional[str])
        - client (Optional [<tusclient.client.TusClient>])
        - chunk_size (Optional[int])
        - metadata (Optional[dict])
        - retries (Optional[int])
        - retry_delay (Optional[int])
        - store_url (Optional[bool])
        - url_storage (Optinal [<tusclient.storage.interface.Storage>])
        - fingerprinter (Optional [<tusclient.fingerprint.interface
        .Fingerprint>])
        - upload_checksum (Optional[bool])
    """
    DEFAULT_HEADERS = {"Tus-Resumable": "1.0.0"}
    DEFAULT_CHUNK_SIZE = MAXSIZE
    CHECKSUM_ALGORITHM_PAIR = ("sha1", hashlib.sha1, )

    def __init__(self, file_path: Optional[str] = None,
                 file_stream: Optional[IO] = None,
                 url: Optional[str] = None,
                 client: Optional['TusClient'] = None,
                 chunk_size: int = MAXSIZE, metadata: Optional[Dict] = None,
                 retries: int = 10, retry_delay: int = 300,
                 store_url=False, upload_checksum=False):
        if file_path is None and file_stream is None:
            raise ValueError(
                "Either 'file_path' or 'file_stream' cannot be None.")

        if url is None and client is None:
            raise ValueError("Either 'url' or 'client' cannot be None.")

        self.file_path = file_path
        self.file_stream = file_stream
        self.stop_at = self.get_file_size()
        self.client = client
        self.metadata = metadata or {}
        self.store_url = store_url
        # self.url_storage = url_storage
        # self.fingerprinter = fingerprinter or Fingerprint()
        self.offset = 0
        self.url = url
        self.__init_url_and_offset(url)
        self.chunk_size = 1048576  # bytes, 1mb
        self.retries = retries
        self.request = None
        self._retried = 0
        self.retry_delay = retry_delay
        self.upload_checksum = upload_checksum
        self.__checksum_algorithm_name, self.__checksum_algorithm = \
            self.CHECKSUM_ALGORITHM_PAIR

    def get_headers(self):
        """
        Return headers of the uploader instance. This would include the headers of the
        client instance.
        """
        client_headers = getattr(self.client, 'headers', {})
        return dict(self.DEFAULT_HEADERS, **client_headers)

    def get_url_creation_headers(self):
        """Return headers required to create upload url"""
        headers = self.get_headers()
        headers['upload-length'] = str(self.get_file_size())
        headers['upload-metadata'] = ','.join(self.encode_metadata())
        return headers

    @property
    def checksum_algorithm(self):
        """The checksum algorithm to be used for the Upload-Checksum extension.
        """
        return self.__checksum_algorithm

    @property
    def checksum_algorithm_name(self):
        """The name of the checksum algorithm to be used for the
        Upload-Checksum
        extension.
        """
        return self.__checksum_algorithm_name

    @catch_requests_error
    def get_offset(self):
        # TODO: make this method async using aiohttp
        """
        Return offset from tus server.

        This is different from the instance attribute 'offset' because this
        makes an
        http request to the tus server to retrieve the offset.
        """
        resp = requests.head(self.url, headers=self.get_headers())
        offset = resp.headers.get('upload-offset')
        if offset is None:
            msg = 'Attempt to retrieve offset fails with status {}'.format(
                resp.status_code)
            raise TusCommunicationError(msg, resp.status_code, resp.content)
        return int(offset)

    def encode_metadata(self):
        """
        Return list of encoded metadata as defined by the Tus protocol.
        """
        encoded_list = []
        for key, value in self.metadata.items():
            key_str = str(key)  # dict keys may be of any object type.

            # confirm that the key does not contain unwanted characters.
            if re.search(r'^$|[\s,]+', key_str):
                msg = 'Upload-metadata key "{}" cannot be empty nor contain spaces or commas.'
                raise ValueError(msg.format(key_str))

            value_bytes = value.encode('latin-1')
            encoded_list.append('{} {}'.format(
                key_str, b64encode(value_bytes).decode('ascii')))
        return encoded_list

    def __init_url_and_offset(self, url: Optional[str] = None):
        """
        Return the tus upload url.
        If resumability is enabled, this would try to get the url from storage
        if available, otherwise it would request a new upload url from the tus server.
        """
        if url:
            self.set_url(url)
        if self.url:
            self.offset = self.get_offset()

    def set_url(self, url: str):
        """Set the upload URL"""
        self.url = url

    def get_request_length(self):
        """
        Return length of next chunk upload.
        """
        remainder = self.stop_at - self.offset
        return self.chunk_size if remainder > self.chunk_size else remainder

    def get_file_stream(self):
        """
        Return a file stream instance of the upload.
        """
        if self.file_stream:
            self.file_stream.seek(0)
            return self.file_stream
        elif os.path.isfile(self.file_path):
            return open(self.file_path, 'rb')
        else:
            raise ValueError("invalid file {}".format(self.file_path))

    def get_file_size(self):
        """
        Return size of the file.
        """
        stream = self.get_file_stream()
        stream.seek(0, os.SEEK_END)
        return stream.tell()



import asyncio
import base64
from functools import wraps
from typing import Optional

import aiohttp
import requests

from .exceptions import TusUploadFailed, TusCommunicationError

# Catches requests exceptions and throws custom tuspy errors.


def catch_requests_error(func):
    """Deocrator to catch requests exceptions"""
    @wraps(func)
    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except requests.exceptions.RequestException as error:
            raise TusCommunicationError(error)

    return _wrapper


class BaseTusRequest:
    """
    Http Request Abstraction.

    Sets up tus custom http request on instantiation.

    requires argument 'uploader' an instance of tusclient.uploader.Uploader
    on instantiation.

    :Attributes:
        - response_headers (dict)
        - file (file):
            The file that is being uploaded.
    """

    def __init__(self, uploader):
        self._url = uploader.url
        self.response_headers = {}
        self.status_code = None
        self.response_content = None
        self.file = uploader.get_file_stream()
        self.file.seek(uploader.offset)

        self._request_headers = {
            'upload-offset': str(uploader.offset),
            'Content-Type': 'application/offset+octet-stream'
        }
        self._request_headers.update(uploader.get_headers())
        self._content_length = uploader.get_request_length()
        self._upload_checksum = uploader.upload_checksum
        self._checksum_algorithm = uploader.checksum_algorithm
        self._checksum_algorithm_name = uploader.checksum_algorithm_name

    def add_checksum(self, chunk: bytes):
        if self._upload_checksum:
            self._request_headers['upload-checksum'] = \
                ' '.join((
                    self._checksum_algorithm_name,
                    base64.b64encode(
                        self._checksum_algorithm(chunk).digest()
                    ).decode('ascii'),
                ))


class TusRequest(BaseTusRequest):
    """Class to handle async Tus upload requests"""
    def perform(self):
        """
        Perform actual request.
        """
        try:
            chunk = self.file.read(self._content_length)
            self.add_checksum(chunk)
            resp = requests.patch(self._url, data=chunk,
                                  headers=self._request_headers)
            self.status_code = resp.status_code
            self.response_content = resp.content
            self.response_headers = {
                k.lower(): v for k, v in resp.headers.items()}
        except requests.exceptions.RequestException as error:
            raise TusUploadFailed(error)


class AsyncTusRequest(BaseTusRequest):
    """Class to handle async Tus upload requests"""
    def __init__(self, *args, io_loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs):
        self.io_loop = io_loop
        super().__init__(*args, **kwargs)

    async def perform(self):
        """
        Perform actual request.
        """

        chunk = self.file.read(self._content_length)
        self.add_checksum(chunk)
        try:
            async with aiohttp.ClientSession(loop=self.io_loop) as session:
                async with session.patch(self._url, data=chunk,
                                         headers=self._request_headers) as resp:
                    self.status_code = resp.status
                    self.response_headers = {
                        k.lower(): v for k, v in resp.headers.items()}
                    self.response_content = await resp.content.read()
        except aiohttp.ClientError as error:
            raise TusUploadFailed(error)



from typing import Dict

from .uploader import Uploader, AsyncUploader


class TusClient:
    """
    Object representation of Tus client for the Backend.AI.

    :Attributes:
        - headers (dict):
            This can be used to set the server specific headers. These headers
            would be sent along with every request made by the cleint to the
            server. This may be used to set
            authentication headers.
            These headers should not include headers required by tus
            protocol. If not set this defaults to an empty dictionary.

    :Constructor Args:
        - headers (Optiional[dict])
    """

    def __init__(self, headers: Dict[str, str] = None):
        self.headers = headers if headers else {}

    def uploader(self, *args, **kwargs) -> Uploader:
        """
        Return uploader instance pointing at current client instance.

        Return uplaoder instance with which you can control the upload of a specific
        file. The current instance of the tus client is passed to the uploader on creation.

        :Args:
            see tusclient.uploader.Uploader for required and optional arguments.
        """
        kwargs['client'] = self
        return Uploader(*args, **kwargs)

    def async_uploader(self, *args, **kwargs) -> AsyncUploader:
        kwargs['client'] = self
        return AsyncUploader(*args, **kwargs)



import asyncio
from urllib.parse import urljoin
import time
from typing import Optional

import requests
from tqdm import tqdm

from .baseuploader import BaseUploader
from .exceptions import TusUploadFailed, TusCommunicationError
from .request import TusRequest, AsyncTusRequest, catch_requests_error


def _verify_upload(request: TusRequest):
    if request.status_code == 204:
        return True
    else:
        raise TusUploadFailed('', request.status_code,
                              request.response_content)


class Uploader(BaseUploader):
    def upload(self, stop_at: Optional[int] = None):
        """
        Perform file upload.

        Performs continous upload of chunks of the file. The size uploaded at
        each cycle is
        the value of the attribute 'chunk_size'.

        :Args:
            - stop_at (Optional[int]):
                Determines at what offset value the upload should stop.
                If not specified this
                defaults to the file size.
        """
        self.stop_at = stop_at or self.get_file_size()

        while self.offset < self.stop_at:
            self.upload_chunk()

    def upload_chunk(self):
        """
        Upload chunk of file.
        """
        self._retried = 0
        if not self.url:
            self.set_url(self.create_url())
            self.offset = 0
        self._do_request()
        self.offset = int(self.request.response_headers.get('upload-offset'))

    @catch_requests_error
    def create_url(self):
        """
        Return upload url.

        Makes request to tus server to create a new upload url for the
        required file upload.
        """
        resp = requests.post(
            self.client.url, headers=self.get_url_creation_headers())
        url = resp.headers.get("location")
        if url is None:
            msg = 'Attempt to retrieve create file url with status {}'.format(
                resp.status_code)
            raise TusCommunicationError(msg, resp.status_code, resp.content)
        return urljoin(self.client.url, url)

    def _do_request(self):
        self.request = TusRequest(self)
        try:
            self.request.perform()
            _verify_upload(self.request)
        except TusUploadFailed as error:
            self._retry_or_cry(error)

    def _retry_or_cry(self, error):
        print("Retry error ", error)
        if self.retries > self._retried:
            time.sleep(self.retry_delay)

            self._retried += 1
            try:
                self.offset = self.get_offset()
            except TusCommunicationError as err:
                self._retry_or_cry(err)
            else:
                self._do_request()
        else:
            raise error


class AsyncUploader(BaseUploader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def upload(self, stop_at: Optional[int] = None):
        """
        Perform file upload.

        Performs continous upload of chunks of the file. The size uploaded at each cycle is
        the value of the attribute 'chunk_size'.

        :Args:
            - stop_at (Optional[int]):
                Determines at what offset value the upload should stop. If not specified this
                defaults to the file size.
        """
        self.stop_at = stop_at or self.get_file_size()
        with tqdm(
            total=self.get_file_size(),
            unit='bytes',
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            while self.offset < self.stop_at:
                await self.upload_chunk()
                pbar.update(self.chunk_size)

    async def upload_chunk(self):
        """
        Upload chunk of file.
        """
        await self._do_request()
        self.offset = int(self.request.response_headers.get('upload-offset'))

    async def _do_request(self):
        self.request = AsyncTusRequest(self)
        try:
            await self.request.perform()
            _verify_upload(self.request)
        except TusUploadFailed as error:
            await self._retry_or_cry(error)

    async def _retry_or_cry(self, error):
        print("Error ", error)
        print("Retries ", self.retries, self._retried)
        if self.retries > self._retried:
            await asyncio.sleep(self.retry_delay)
            self._retried += 1
            try:
                self.offset = self.get_offset()
            except TusCommunicationError as err:
                await self._retry_or_cry(err)
            else:
                await self._do_request()
        else:
            print("Error ", error)
            raise error



from setuptools import setup
from pathlib import Path
import re


setup_requires = [
    'setuptools>=49.6.0',
]

install_requires = [
    'aiohttp>=3.6.2',
    'tqdm>=4.42'
]
build_requires = [
    'wheel>=0.35.1',
    'twine>=3.1.1',
    'towncrier>=19.2.0',
]
test_requires = [
    'pytest~=6.0.1',
    'pytest-cov',
    'pytest-mock',
    'pytest-asyncio>=0.14.0',
    'aioresponses~=0.6.3',
    'codecov',
]
lint_requires = [
    'flake8>=3.8.1',
]
typecheck_requires = [
    'mypy>=0.782',
]


def get_version():
    src = (Path(__file__).parent / 'aiotusclient' / '__init__.py').read_text()
    m = re.search(r"""^__version__\s*=\s*(["'])([^'"]+)\1$""", src)
    return m.group(2)


setup(
    name='aiotusclient',
    version=get_version(),
    description='tus.io-compatible upload client library for Python asyncio',
    url='https://github.com/lablup/aiotusclient',
    author='Lablup Inc.',
    author_email='sergey@lablup.com',
    license='MIT',
    packages=['aiotusclient'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
    ],
    python_requires='>=3.7',
    setup_requires=setup_requires,
    install_requires=install_requires,
    extras_require={
        'build': build_requires,
        'test': test_requires,
        'lint': lint_requires,
        'typecheck': typecheck_requires,
    },
)
